# This file is part of tad-dftd4.
#
# SPDX-Identifier: Apache-2.0
# Copyright (C) 2024 Grimme Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Dispersion Methods: Base
========================

Base classes and interfaces for dispersion terms.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import ClassVar

import torch
from tad_mctc.typing import DD, Any, CNFunction, Tensor, TensorLike

from ..cutoff import Cutoff
from ..damping import Damping, Param
from ..model import ModelInst, ModelKey


class DispTerm(TensorLike, ABC):
    """
    Base class for all dispersion terms.

    Parameters
    ----------
    damping_fn : Damping
        Damping function to be used for the dispersion term.
    charge_dependent : bool
        Whether the term is charge-dependent, i.e., requires atomic charges
        for the calculation.
    device : torch.device, optional
        Device on which the term is calculated.
    dtype : torch.dtype, optional
        Data type of the term's tensors.
    """

    damping_fn: Damping
    """Damping function to be used for the dispersion term."""

    charge_dependent: bool
    """
    Whether the term is charge-dependent, i.e., requires atomic charges for
    the calculation.
    """

    __slots__ = ("damping_fn", "charge_dependent")

    def __init__(
        self,
        damping_fn: Damping,
        charge_dependent: bool,
        *,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ):
        super().__init__(device=device, dtype=dtype)

        self.damping_fn = damping_fn
        self.charge_dependent = charge_dependent

    def __eq__(self, other: Any):
        if self.__class__ is not other.__class__:
            return False

        return (
            self.damping_fn == other.damping_fn
            and self.charge_dependent == other.charge_dependent
        )

    @abstractmethod
    def calculate(
        self,
        numbers: Tensor,
        positions: Tensor,
        param: Param,
        cn: Tensor,
        model: ModelInst,
        q: Tensor | None,
        r4r2: Tensor,
        rvdw: Tensor,
        cutoff: Cutoff,
    ) -> Tensor:
        """Evaluate the energy for the dispersion term."""


class Disp(TensorLike):
    """Base class for DFT-D dispersion calculations."""

    terms: list[DispTerm]
    """List of dispersion terms for which the calculation is performed."""

    cn_fn: CNFunction
    """Coordination number."""

    cn_fn_kwargs: dict[str, Any]
    """Keyword arguments for the coordination number function."""

    _model_key: ModelKey
    """Key for the DFT-D model, e.g., 'd4'."""

    _model_kwargs: dict[str, Any]
    """Keyword arguments for the DFT-D model."""

    _model_instance: ModelInst | None
    """Instance of the DFT-D model, if provided."""

    __slots__ = (
        "terms",
        "cn_fn",
        "cn_fn_kwargs",
        "_model_key",
        "_model_instance",
        "_model_kwargs",
    )

    _ALLOWED_MODELS = ("d3", "d4", "d4s", "d5")
    """Allowed DFT-D models for the calculation."""

    TERMS: ClassVar[list[tuple[type[DispTerm], dict[str, Any] | None]]] = []
    """List of dispersion terms to be registered in the constructor."""

    def __init__(
        self,
        model: ModelKey | ModelInst = "d4",
        model_kwargs: dict[str, Any] | None = None,
        cn_fn: CNFunction | None = None,
        cn_fn_kwargs: dict[str, Any] | None = None,
        *,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ):
        super().__init__(device=device, dtype=dtype)

        if isinstance(model, str):
            key = model.casefold()

            if key not in self._ALLOWED_MODELS:
                raise ValueError(
                    f"Unknown model '{key}'. "
                    f"Please use {', '.join(self._ALLOWED_MODELS)}."
                )

            self._model_key = key
            self._model_instance = None
            self._model_kwargs = model_kwargs or {}

        else:
            # sentinel, never queried
            self._model_key = "instance"  # type: ignore[assignment]

            self._model_instance = model
            self._model_kwargs = {}
            if model_kwargs:
                from warnings import warn

                warn(
                    "`model` is an instance - `model_kwargs` were ignored.",
                    RuntimeWarning,
                    stacklevel=2,
                )

        if cn_fn is None:
            name = (
                self._model_instance.__class__.__name__
                if self._model_instance is not None
                else self._model_key
            )
            if "3" in name:
                # pylint: disable=import-outside-toplevel
                from tad_mctc.ncoord import cn_d3

                cn_fn = cn_d3
            elif "4" in name:
                # pylint: disable=import-outside-toplevel
                from tad_mctc.ncoord import cn_d4

                cn_fn = cn_d4
            elif "5" in name:
                # pylint: disable=import-outside-toplevel
                from tad_mctc.ncoord import cn_d3

                cn_fn = cn_d3
            else:  # pragma: no cover
                raise ValueError(
                    f"Unknown model '{self._model_key}'. "
                    "Please use 'd3', 'd4', 'd4s', or 'd5'."
                )

        self.cn_fn = cn_fn
        self.cn_fn_kwargs = cn_fn_kwargs if cn_fn_kwargs is not None else {}

        self.terms: list[DispTerm] = []
        for term_cls, kw in self.TERMS:
            kw = {} if kw is None else kw.copy()
            self.register(term_cls(**kw))

    def register(self, term: DispTerm) -> None:
        self.terms.append(term)

    def deregister(self, term: DispTerm) -> None:
        self.terms.remove(term)

    def get_model(self, numbers: Tensor) -> ModelInst:
        """
        Get the DFT-D4 model for the given atomic numbers.

        Parameters
        ----------
        numbers : Tensor
            Atomic numbers of the atoms in the system.

        Returns
        -------
        ModelInst
            The DFT-D4 model initialized with the atomic numbers.
        """
        if self._model_instance is not None:
            return self._model_instance

        if self._model_key.casefold() == "d3":
            # pylint: disable=import-outside-toplevel
            from tad_dftd4.model.d3 import D3Model

            return D3Model(numbers=numbers, **self._model_kwargs, **self.dd)

        if self._model_key.casefold() == "d4":
            # pylint: disable=import-outside-toplevel
            from tad_dftd4.model.d4 import D4Model

            return D4Model(numbers=numbers, **self._model_kwargs, **self.dd)

        if self._model_key.casefold() == "d4s":
            # pylint: disable=import-outside-toplevel
            from tad_dftd4.model.d4s import D4SModel

            return D4SModel(numbers=numbers, **self._model_kwargs, **self.dd)

        raise ValueError(
            f"Unknown model '{self._model_key}'. "
            "Please use 'd3', 'd4', 'd4s', or 'd5'."
        )

    # Radii

    def get_rcov(self, numbers: Tensor) -> Tensor:
        # pylint: disable=import-outside-toplevel
        from tad_mctc.data import COV_D3

        return COV_D3(**self.dd)[numbers]

    def get_r4r2(self, numbers: Tensor) -> Tensor:
        # pylint: disable=import-outside-toplevel
        from tad_dftd4.data import R4R2

        return R4R2(**self.dd)[numbers]

    def get_rvdw(self, numbers: Tensor) -> Tensor:
        # pylint: disable=import-outside-toplevel
        from tad_mctc.data import VDW_PAIRWISE

        return VDW_PAIRWISE(**self.dd)[
            numbers.unsqueeze(-1), numbers.unsqueeze(-2)
        ]

    # Calculation

    def calculate(
        self,
        numbers: Tensor,
        positions: Tensor,
        charge: Tensor,
        param: Param,
        *,
        cutoff: Cutoff | None = None,
        q: Tensor | None = None,
        rcov: Tensor | None = None,
        r4r2: Tensor | None = None,
        rvdw: Tensor | None = None,
    ):
        """
        Evaluate DFT-D4 dispersion energy for a (batch of) molecule(s).

        Parameters
        ----------
        numbers : Tensor
            Atomic numbers for all atoms in the system of shape ``(..., nat)``.
        positions : Tensor
            Cartesian coordinates of all atoms (shape: ``(..., nat, 3)``).
        charge : Tensor
            Total charge of the system.
        param : Param
            DFT-D4 damping parameters.
        model : D4Model | D4SModel | None, optional
            The DFT-D4 dispersion model for the evaluation of the C6
            coefficients. Defaults to ``None``, which creates
            :class:`tad_dftd4.model.d4.D4Model`.
        rcov : Tensor | None, optional
            Covalent radii of the atoms in the system. Defaults to
            ``None``, i.e., default values are used.
        r4r2 : Tensor | None, optional
            r⁴ over r² expectation values of the atoms in the system. Defaults
            to ``None``, i.e., default values are used.
        q : Tensor | None, optional
            Atomic partial charges. Defaults to ``None``, i.e., EEQ charges are
            calculated using the total ``charge``.
        cutoff : Cutoff | None, optional
            Collection of real-space cutoffs. Defaults to ``None``, i.e.,
            :class:`tad_dftd4.cutoff.Cutoff` is initialized with its defaults.
        counting_function : CountingFunction, optional
            Counting function used for the DFT-D4 coordination number. Defaults
            to the error function counting function
            :func:`tad_mctc.ncoord.count.erf_count`.
        damping_function : DampingFunction, optional
            Damping function to evaluate distance dependent contributions.
            Defaults to the Becke-Johnson rational damping function
            :func:`tad_dftd4.damping.rational.rational_damping`.

        Returns
        -------
        Tensor
            Atom-resolved DFT-D4 dispersion energy.

        Raises
        ------
        ValueError
            Shape inconsistencies between ``numbers``, ``positions``, ``r4r2``,
            or, ``rcov``.
        RuntimeError
            If atomic charges are explicitly provided, but no term requires
            them.
        """
        dd: DD = {"device": positions.device, "dtype": positions.dtype}

        if numbers.shape != positions.shape[:-1]:
            raise ValueError(
                f"Shape of positions ({positions.shape}) is not consistent "
                f"with atomic numbers ({numbers.shape}).",
            )

        if cutoff is None:
            cutoff = Cutoff(**dd)

        model = self.get_model(numbers=numbers)

        # 2) radii defaults
        if r4r2 is None:
            r4r2 = self.get_r4r2(numbers)
        if numbers.shape != r4r2.shape:
            raise ValueError(
                f"Shape of expectation values r4r2 ({r4r2.shape}) is not "
                f"consistent with atomic numbers ({numbers.shape}).",
            )

        if rcov is None:
            rcov = self.get_rcov(numbers)
        if numbers.shape != rcov.shape:
            raise ValueError(
                f"Shape of covalent radii ({rcov.shape}) is not consistent "
                f"with atomic numbers ({numbers.shape}).",
            )

        if rvdw is None:
            rvdw = self.get_rvdw(numbers)
        if numbers.shape != rvdw.shape[:-1]:
            raise ValueError(
                f"Shape of van der Waals radii ({rvdw.shape}) is not "
                f"consistent with atomic numbers ({numbers.shape}).",
            )

        # 3) Coordination numbers
        cn = self.cn_fn(
            numbers, positions, rcov=rcov, cutoff=cutoff.cn, **self.cn_fn_kwargs
        )

        # 4) charges if any term demands them
        is_c_dep = any(t.charge_dependent for t in self.terms)
        if q is not None and is_c_dep is False:
            raise RuntimeError(
                "Atomic charges are explicitly provided, but no term "
                "requires them. Please remove the `q` argument or "
                "provide a term that requires atomic charges.",
            )

        if q is None and is_c_dep is True:
            # pylint: disable=import-outside-toplevel
            from tad_multicharge import get_eeq_charges

            q = get_eeq_charges(
                numbers, positions, charge, cutoff=cutoff.cn_eeq
            )

        if q is not None:
            if numbers.shape != q.shape:
                raise ValueError(
                    f"Shape of atomic charges ({q.shape}) is not consistent "
                    f"with atomic numbers ({numbers.shape}).",
                )

        # 5) delegate
        energy = torch.zeros_like(numbers, **dd)
        for term in self.terms:
            energy = energy + term.calculate(
                numbers=numbers,
                positions=positions,
                param=param,
                cn=cn,
                model=model,
                q=q,
                r4r2=r4r2,
                rvdw=rvdw,
                cutoff=cutoff,
            )

        return energy



# This file is part of tad-dftd4.
#
# SPDX-Identifier: Apache-2.0
# Copyright (C) 2024 Grimme Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Data: Expectation values
========================

PBE0/def2-QZVP atomic values calculated by S. Grimme in Gaussian (2010).
Rare gases recalculated by J. Mewes with PBE0/aug-cc-pVQZ in Dirac (2018).
Also new super heavies Cn, Nh, Fl, Lv, Og and Am-Rg calculated at
4c-PBE/Dyall-AE4Z level (Dirac 2022).
"""
from __future__ import annotations

import torch

__all__ = ["R4R2"]


def R4R2(
    dtype: torch.dtype | None = None, device: torch.device | None = None
) -> torch.Tensor:
    """
    Returns the r⁴ over r² expectation values as a tensor.

    Parameters
    ----------
    dtype : torch.dtype | None, optional
        The desired data type of the returned tensor. Defaults to None.
    device : torch.device | None, optional
        The desired device of the returned tensor. Defaults to None.

    Returns
    -------
    Tensor
        A tensor containing the r⁴ over r² expectation values.
    """

    # Actually calculated r⁴ over r² expectation values.
    # fmt: off
    _r4_over_r2 = [
        0.0000,  # None
        8.0589, 3.4698,  # H,He
        29.0974,14.8517,11.8799, 7.8715, 5.5588, 4.7566, 3.8025, 3.1036,  # Li-Ne
        26.1552,17.2304,17.7210,12.7442, 9.5361, 8.1652, 6.7463, 5.6004,  # Na-Ar
        29.2012,22.3934,  # K,Ca
        19.0598,16.8590,15.4023,12.5589,13.4788,  # Sc-
        12.2309,11.2809,10.5569,10.1428, 9.4907,  # -Zn
        13.4606,10.8544, 8.9386, 8.1350, 7.1251, 6.1971,  # Ga-Kr
        30.0162,24.4103,  # Rb,Sr
        20.3537,17.4780,13.5528,11.8451,11.0355,  # Y-
        10.1997, 9.5414, 9.0061, 8.6417, 8.9975,  # -Cd
        14.0834,11.8333,10.0179, 9.3844, 8.4110, 7.5152,  # In-Xe
        32.7622,27.5708,  # Cs,Ba
        23.1671,21.6003,20.9615,20.4562,20.1010,19.7475,19.4828,  # La-Eu
        15.6013,19.2362,17.4717,17.8321,17.4237,17.1954,17.1631,  # Gd-Yb
        14.5716,15.8758,13.8989,12.4834,11.4421,  # Lu-
        10.2671, 8.3549, 7.8496, 7.3278, 7.4820,  # -Hg
        13.5124,11.6554,10.0959, 9.7340, 8.8584, 8.0125,  # Tl-Rn
        29.8135,26.3157,  # Fr,Ra
        19.1885,15.8542,16.1305,15.6161,15.1226,16.1576,14.6510,  # Ac-Am
        14.7178,13.9108,13.5623,13.2326,12.9189,12.6133,12.3142,  # Cm-No
        14.8326,12.3771,10.6378, 9.3638, 8.2297,  # Lr-
        7.5667, 6.9456, 6.3946, 5.9159, 5.4929,  # -Cn
        6.7286, 6.5144,10.9169,10.3600, 9.4723, 8.6641,  # Nh-Og
    ]
    # fmt: on

    sqrtz = torch.sqrt(
        torch.arange(len(_r4_over_r2), device=device, dtype=dtype)
    )
    return torch.sqrt(
        0.5 * (torch.tensor(_r4_over_r2, device=device, dtype=dtype) * sqrtz)
    )



# This file is part of tad-dftd4.
#
# SPDX-Identifier: Apache-2.0
# Copyright (C) 2024 Grimme Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utils
=====

Utility functions for this project.
"""
from __future__ import annotations

import torch
from tad_mctc.math import einsum
from tad_mctc.typing import Tensor

__all__ = ["trapzd", "trapzd_atm", "trapzd_noref", "is_exceptional"]


def trapzd(pol1: Tensor, pol2: Tensor | None = None) -> Tensor:
    """
    Numerical Casimir--Polder integration.

    Parameters
    ----------
    pol1 : Tensor
        Polarizabilities of shape ``(..., nat, nref, 23)``.
    pol2 : Tensor | None, optional
        Polarizabilities of shape ``(..., nat, nref, 23)``. Defaults to
        ``None``, in which case ``pol2`` is set to ``pol1``.

    Returns
    -------
    Tensor
        C6 coefficients of shape ``(..., nat, nat, nref, nref)``.
    """
    thopi = 3.0 / 3.141592653589793238462643383279502884197

    weights = torch.tensor(
        [
            2.4999500000000000e-002,
            4.9999500000000000e-002,
            7.5000000000000010e-002,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1500000000000000,
            0.2000000000000000,
            0.2000000000000000,
            0.2000000000000000,
            0.2000000000000000,
            0.3500000000000000,
            0.5000000000000000,
            0.7500000000000000,
            1.0000000000000000,
            1.7500000000000000,
            2.5000000000000000,
            1.2500000000000000,
        ],
        device=pol1.device,
        dtype=pol1.dtype,
    )

    # NOTE: In the old version, a memory inefficient intermediate tensor was
    # created. The new version uses `einsum` to avoid this.
    #
    # (..., 1, nat, 1, nref, 23) * (..., nat, 1, nref, 1, 23) =
    # (..., nat, nat, nref, nref, 23) -> (..., nat, nat, nref, nref)
    # a = alpha.unsqueeze(-4).unsqueeze(-3) * alpha.unsqueeze(-3).unsqueeze(-2)
    #
    # rc6 = thopi * torch.sum(weights * a, dim=-1)

    return thopi * einsum(
        "w,...iaw,...jbw->...ijab",
        *(weights, pol1, pol1 if pol2 is None else pol2),
    )


def trapzd_noref(pol1: Tensor, pol2: Tensor | None = None) -> Tensor:
    """
    Numerical Casimir--Polder integration.

    This version takes polarizabilities of shape ``(..., nat, 23)``, i.e.,
    the reference dimension has already been summed over.

    Parameters
    ----------
    pol1 : Tensor
        Polarizabilities of shape ``(..., nat, 23)``.
    pol2 : Tensor | None, optional
        Polarizabilities of shape ``(..., nat, 23)``. Defaults to
        ``None``, in which case ``pol2`` is set to ``pol1``.

    Returns
    -------
    Tensor
        C6 coefficients of shape ``(..., nat, nat)``.
    """
    thopi = 3.0 / 3.141592653589793238462643383279502884197

    weights = torch.tensor(
        [
            2.4999500000000000e-002,
            4.9999500000000000e-002,
            7.5000000000000010e-002,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1500000000000000,
            0.2000000000000000,
            0.2000000000000000,
            0.2000000000000000,
            0.2000000000000000,
            0.3500000000000000,
            0.5000000000000000,
            0.7500000000000000,
            1.0000000000000000,
            1.7500000000000000,
            2.5000000000000000,
            1.2500000000000000,
        ],
        device=pol1.device,
        dtype=pol1.dtype,
    )

    return thopi * einsum(
        "w,...iw,...jw->...ij",
        *(weights, pol1, pol1 if pol2 is None else pol2),
    )


def trapzd_atm(pol: Tensor) -> Tensor:
    """
    Numerical Casimir--Polder integration for ATM term.

    This version takes polarizabilities of shape
    ``(..., nat, nat, nat, 23)``, i.e., the reference dimension has
    already been summed over.

    Parameters
    ----------
    pol : Tensor
        Polarizabilities of shape ``(..., nat, nat, nat, 23)``.

    Returns
    -------
    Tensor
        C9 coefficients of shape ``(..., nat, nat, nat)``.
    """
    thopi = 3.0 / 3.141592653589793238462643383279502884197

    weights = torch.tensor(
        [
            2.4999500000000000e-002,
            4.9999500000000000e-002,
            7.5000000000000010e-002,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1000000000000000,
            0.1500000000000000,
            0.2000000000000000,
            0.2000000000000000,
            0.2000000000000000,
            0.2000000000000000,
            0.3500000000000000,
            0.5000000000000000,
            0.7500000000000000,
            1.0000000000000000,
            1.7500000000000000,
            2.5000000000000000,
            1.2500000000000000,
        ],
        device=pol.device,
        dtype=pol.dtype,
    )

    return thopi * einsum("...ijkw,w->...ijk", pol, weights)


def is_exceptional(x: Tensor, dtype: torch.dtype) -> Tensor:
    """
    Check if a tensor is exceptional (``NaN`` or too large).

    Parameters
    ----------
    x : Tensor
        Tensor to check.
    dtype : torch.dtype
        Data type of the tensor.

    Returns
    -------
    Tensor
        Boolean tensor indicating exceptional values.
    """
    return torch.isnan(x) | (x > torch.finfo(dtype).max)



# This file is part of tad-dftd4.
#
# SPDX-Identifier: Apache-2.0
# Copyright (C) 2024 Grimme Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Model: Typing
=============

Type annotations for models.
"""
from __future__ import annotations

from typing import Union

from tad_mctc.typing import Literal, TypeAlias

from .d3 import D3Model
from .d4 import D4Model
from .d4s import D4SModel
from .d5 import D5Model

__all__ = ["ModelKey", "ModelInst", "ModelInstD3", "ModelInstD4"]


ModelKey: TypeAlias = Literal["d3", "d4", "d4s", "d5"]

ModelInstD3: TypeAlias = D3Model
ModelInstD4: TypeAlias = Union[D4Model, D4SModel]
ModelInstD5: TypeAlias = D5Model

ModelInst: TypeAlias = Union[ModelInstD3, ModelInstD4, ModelInstD5]



# This file is part of tad-dftd4.
#
# SPDX-Identifier: Apache-2.0
# Copyright (C) 2024 Grimme Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Dispersion energy
=================

This module provides the dispersion energy evaluation for the pairwise
interactions. It contains the main entrypoint for the dispersion energy
(:func:`.dftd4`).
"""
from __future__ import annotations

import torch
from tad_mctc.ncoord import cn_d4, erf_count
from tad_mctc.typing import DD, CNFunction, CountingFunction, Tensor
from tad_multicharge import get_eeq_charges

from .cutoff import Cutoff
from .damping import Damping, Param, RationalDamping, ZeroDamping
from .dispersion import Disp
from .dispersion.d4 import D4ATMApprox
from .dispersion.twobody import TwoBodyTerm
from .model import D4Model, ModelInst, ModelKey

__all__ = ["dftd4", "get_properties"]


def dftd4(
    numbers: Tensor,
    positions: Tensor,
    charge: Tensor,
    param: Param,
    *,
    model: ModelKey | ModelInst = "d4",
    rcov: Tensor | None = None,
    r4r2: Tensor | None = None,
    rvdw: Tensor | None = None,
    q: Tensor | None = None,
    cutoff: Cutoff | None = None,
    cn_function: CNFunction = cn_d4,
    counting_function: CountingFunction = erf_count,
    damping_function: Damping = RationalDamping(),
) -> Tensor:
    """
    Evaluate DFT-D4 dispersion energy for a (batch of) molecule(s).

    Parameters
    ----------
    numbers : Tensor
        Atomic numbers for all atoms in the system of shape ``(..., nat)``.
    positions : Tensor
        Cartesian coordinates of all atoms (shape: ``(..., nat, 3)``).
    charge : Tensor
        Total charge of the system.
    param : Param
        DFT-D4 damping parameters.
    model : D4Model | D4SModel | None, optional
        The DFT-D4 dispersion model for the evaluation of the C6 coefficients.
        Defaults to ``None``, which creates :class:`tad_dftd4.model.d4.D4Model`.
    rcov : Tensor | None, optional
        Covalent radii of the atoms in the system. Defaults to
        ``None``, i.e., default values are used.
    r4r2 : Tensor | None, optional
        r⁴ over r² expectation values of the atoms in the system. Defaults to
        ``None``, i.e., default values are used.
    rvdw : Tensor | None, optional
        Pairwise van der Waals radii of the atoms in the system. Defaults to
        ``None``, i.e., default values are used.
    q : Tensor | None, optional
        Atomic partial charges. Defaults to ``None``, i.e., EEQ charges are
        calculated using the total ``charge``.
    cutoff : Cutoff | None, optional
        Collection of real-space cutoffs. Defaults to ``None``, i.e.,
        :class:`tad_dftd4.cutoff.Cutoff` is initialized with its defaults.
    cn_function : CNFunction, optional
        Function to calculate the coordination number. Defaults to
        :func:`tad_mctc.ncoord.cn_d4`, which uses the
        :func:`tad_mctc.ncoord.count.erf_count` counting function.
    counting_function : CountingFunction, optional
        Counting function used for the DFT-D4 coordination number. Defaults to
        the error function counting function
        :func:`tad_mctc.ncoord.count.erf_count`.
    damping_function : DampingFunction, optional
        Damping function to evaluate distance dependent contributions. Defaults
        to the Becke-Johnson rational damping function
        :func:`tad_dftd4.damping.rational.rational_damping`.

    Returns
    -------
    Tensor
        Atom-resolved DFT-D4 dispersion energy.

    Raises
    ------
    ValueError
        Shape inconsistencies between ``numbers``, ``positions``, ``r4r2``,
        or, ``rcov``.
    """
    dd: DD = {"device": positions.device, "dtype": positions.dtype}

    disp = Disp(
        cn_fn=cn_function,
        cn_fn_kwargs={"counting_function": counting_function},
        model=model,
        **dd,
    )

    twobody_term = TwoBodyTerm(
        damping_fn=damping_function,
        charge_dependent=True,
    )
    disp.register(twobody_term)

    threebody_term = D4ATMApprox(
        damping_fn=ZeroDamping(),
        charge_dependent=False,
    )
    disp.register(threebody_term)

    return disp.calculate(
        numbers,
        positions,
        charge,
        param,
        cutoff=cutoff,
        q=q,
        rcov=rcov,
        r4r2=r4r2,
        rvdw=rvdw,
    )


def get_properties(
    numbers: Tensor,
    positions: Tensor,
    charge: Tensor | None = None,
    cutoff: Cutoff | None = None,
) -> tuple[Tensor, Tensor, Tensor, Tensor]:
    """
    Wrapper to evaluate properties related to this dispersion model.

    Parameters
    ----------
    numbers : Tensor
        Atomic numbers for all atoms in the system of shape ``(..., nat)``.
    positions : Tensor
        Cartesian coordinates of all atoms (shape: ``(..., nat, 3)``).
    charge : Tensor | None
        Total charge of the system. Defaults to ``None``, i.e., ``0``.
    cutoff : Cutoff | None
        Real-space cutoff. Defaults to ``None``, i.e., the defaults for
        all cutoffs are used.

    Returns
    -------
    (Tensor, Tensor, Tensor, Tensor)
        Properties related to the dispersion model:
        - DFT-D4 coordination number
        - Atomic partial charges
        - Atom-resolved C6 dispersion coefficients
        - Static polarizabilities
    """
    dd: DD = {"device": positions.device, "dtype": positions.dtype}

    if cutoff is None:
        cutoff = Cutoff(**dd)

    if charge is None:
        charge = torch.tensor(0.0, **dd)

    cn = cn_d4(numbers, positions, cutoff=cutoff.cn)
    q = get_eeq_charges(numbers, positions, charge, cutoff=cutoff.cn_eeq)

    model = D4Model(numbers, **dd)
    weights = model.weight_references(cn, q)
    c6 = model.get_atomic_c6(weights)
    alpha = model.get_polarizabilities(weights)

    return cn, q, c6, alpha



# This file is part of tad-dftd4.
#
# SPDX-Identifier: Apache-2.0
# Copyright (C) 2024 Grimme Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Model: D4S
==========

This module contains the definition of the D4 dispersion model for the
evaluation of C6 coefficients.

Upon instantiation, the reference polarizabilities are calculated for the
all atoms of the molecule(s) and stored in the model class.


Example
-------
>>> import torch
>>> import tad_dftd4 as d4
>>>
>>> numbers = torch.tensor([14, 1, 1, 1, 1]) # SiH4
>>> model = d4.D4Model(numbers)
>>>
>>> # calculate Gaussian weights, optionally pass CN and partial charges
>>> gw = model.weight_references()
>>> c6 = model.get_atomic_c6(gw)
"""
from __future__ import annotations

import torch
from tad_mctc.math import einsum
from tad_mctc.typing import Literal, Tensor, overload

from .. import data
from ..utils import is_exceptional
from .base import WF_DEFAULT, BaseModel

__all__ = ["D4Model"]


class D4Model(BaseModel):
    """
    The D4 dispersion model.
    """

    def _get_wf(self) -> Tensor:
        """Default weighting factor."""
        return torch.tensor(WF_DEFAULT, **self.dd)

    @overload
    def weight_references(
        self,
        cn: Tensor | None = None,
        q: Tensor | None = None,
        *,
        with_dgwdq: Literal[False] = ...,
        with_dgwdcn: Literal[False] = ...,
    ) -> Tensor: ...

    @overload
    def weight_references(
        self,
        cn: Tensor | None = None,
        q: Tensor | None = None,
        *,
        with_dgwdq: Literal[True],
        with_dgwdcn: Literal[False] = ...,
    ) -> tuple[Tensor, Tensor]: ...

    @overload
    def weight_references(
        self,
        cn: Tensor | None = None,
        q: Tensor | None = None,
        *,
        with_dgwdq: Literal[False] = ...,
        with_dgwdcn: Literal[True],
    ) -> tuple[Tensor, Tensor]: ...

    @overload
    def weight_references(
        self,
        cn: Tensor | None = None,
        q: Tensor | None = None,
        *,
        with_dgwdq: Literal[True],
        with_dgwdcn: Literal[True],
    ) -> tuple[Tensor, Tensor, Tensor]: ...

    def weight_references(
        self,
        cn: Tensor | None = None,
        q: Tensor | None = None,
        *,
        with_dgwdq: bool = False,
        with_dgwdcn: bool = False,
    ) -> Tensor | tuple[Tensor, Tensor] | tuple[Tensor, Tensor, Tensor]:
        """
        Calculate the weights of the reference system
        (shape: ``(..., nat, nref)``).

        Parameters
        ----------
        cn : Tensor | None, optional
            Coordination number of every atom. Defaults to `None` (0).
        q : Tensor | None, optional
            Partial charge of every atom. Defaults to `None` (0).
        with_dgwdq : bool, optional
            Whether to also calculate the derivative of the weights with
            respect to the partial charges. Defaults to `False`.
        with_dgwdcn : bool, optional
            Whether to also calculate the derivative of the weights with
            respect to the coordination numbers. Defaults to `False`.

        Returns
        -------
        Tensor | tuple[Tensor, Tensor] | tuple[Tensor, Tensor, Tensor]
            Weights for the atomic reference systems (shape:
            ``(..., nat, ref)``). If ``with_dgwdq`` is ``True``, also returns
            the derivative of the weights with respect to the partial charges.
            If ``with_dgwdcn`` is ``True``, also returns the derivative of the
            weights with respect to the coordination numbers.
        """
        if cn is None:
            cn = torch.zeros(self.numbers.shape, **self.dd)
        if q is None:
            q = torch.zeros(self.numbers.shape, **self.dd)

        if self.ref_charges == "eeq":
            # pylint: disable=import-outside-toplevel
            from ..reference.d4.charge_eeq import clsq as _refq
        elif self.ref_charges == "gfn2":
            # pylint: disable=import-outside-toplevel
            from ..reference.d4.charge_gfn2 import refq as _refq
        else:
            raise ValueError(f"Unknown reference charges: {self.ref_charges}")

        # pylint: disable=import-outside-toplevel
        from ..reference import d4 as d4ref

        refq = _refq.to(**self.dd)[self.numbers]

        zero = torch.tensor(0.0, **self.dd)
        zero_double = torch.tensor(0.0, device=self.device, dtype=torch.double)

        refc = d4ref.refc.to(self.device)[self.numbers]
        mask = refc > 0

        # Due to the exponentiation, `norm` and `expw` may become very small
        # (down to 1e-300). This causes problems for the division by `norm`,
        # since single precision, i.e. `torch.float`, only goes to around 1e-38.
        # Consequently, some values become zero although the actual result
        # should be close to one. The problem does not arise when using `torch.
        # double`. In order to avoid this error, which is also difficult to
        # detect, this part always uses `torch.double`. `params.refcovcn` is
        # saved with `torch.double`, but I still made sure...
        refcn = d4ref.refcovcn.to(device=self.device, dtype=torch.double)[
            self.numbers
        ]

        # For vectorization, we reformulate the Gaussian weighting function:
        # exp(-wf * igw * (cn - cn_ref)^2) = [exp(-(cn - cn_ref)^2)]^(wf *igw)
        # Gaussian weighting function part 1: exp(-(cn - cn_ref)^2)
        dcn = cn.unsqueeze(-1).type(torch.double) - refcn
        tmp = torch.exp(-dcn * dcn)

        # Gaussian weighting function part 2: tmp^(wf * igw)
        # (While the Fortran version just loops over the number of gaussian
        # weights `igw`, we have to use masks and explicitly implement the
        # formulas for exponentiation. Luckily, `igw` only takes on the values
        # 1 and 3.)
        def refc_pow(n: int) -> Tensor:
            return sum(
                (torch.pow(tmp, i * self.wf) for i in range(1, n + 1)),
                torch.tensor(0.0, device=tmp.device),
            )

        refc_pow_1 = torch.where(refc == 1, refc_pow(1), tmp)
        refc_pow_final = torch.where(refc == 3, refc_pow(3), refc_pow_1)

        expw = torch.where(mask, refc_pow_final, zero_double)

        # Normalize weights, but keep shape. This needs double precision.
        # Moreover, we need to mask the normalization to avoid division by zero
        # for autograd. Strangely, `storch.divide` gives erroneous results for
        # some elements (Mg, e.g. in MB16_43/03).
        norm = torch.where(
            mask,
            torch.sum(expw, dim=-1, keepdim=True),
            torch.tensor(1e-300, device=self.device, dtype=torch.double),
        )

        # back to real dtype
        gw_temp = (expw / norm).type(self.dtype)

        # maximum reference CN for each atom
        maxcn = torch.max(refcn, dim=-1, keepdim=True)[0]

        # prevent division by 0 and small values
        gw = torch.where(
            is_exceptional(gw_temp, self.dtype),
            torch.where(refcn == maxcn, torch.tensor(1.0, **self.dd), zero),
            gw_temp,
        )

        # unsqueeze for reference dimension
        zeff = data.ZEFF(self.device)[self.numbers].unsqueeze(-1)
        gam = data.GAM(**self.dd)[self.numbers].unsqueeze(-1) * self.gc
        q = q.unsqueeze(-1)

        # charge scaling
        zeta = torch.where(mask, self._zeta(gam, refq + zeff, q + zeff), zero)

        if with_dgwdq is False and with_dgwdcn is False:
            return zeta * gw

        # DERIVATIVES

        outputs = [zeta * gw]

        if with_dgwdcn is True:

            def _dpow(n: int) -> Tensor:
                return sum(
                    (
                        2 * i * self.wf * dcn * torch.pow(tmp, i * self.wf)
                        for i in range(1, n + 1)
                    ),
                    zero_double,
                )

            wf_1 = torch.where(refc == 1, _dpow(1), zero_double)
            wf_3 = torch.where(refc == 3, _dpow(3), zero_double)
            dexpw = wf_1 + wf_3

            # no mask needed here, already masked in `dexpw`
            dnorm = torch.sum(dexpw, dim=-1, keepdim=True)

            _dgw = (dexpw - expw * dnorm / norm) / norm
            dgw = torch.where(
                is_exceptional(_dgw, self.dtype), zero, _dgw.type(self.dtype)
            )

            outputs.append(zeta * dgw)

        if with_dgwdq is True:
            dzeta = torch.where(
                mask, self._dzeta(gam, refq + zeff, q + zeff), zero
            )

            outputs.append(dzeta * gw)

        return tuple(outputs)  # type: ignore

    def get_atomic_c6(self, gw: Tensor) -> Tensor:
        """
        Calculate atomic C6 dispersion coefficients.

        Parameters
        ----------
        gw : Tensor
            Weights for the atomic reference systems of shape
            `(..., nat, nref)`.

        Returns
        -------
        Tensor
            C6 coefficients for all atom pairs of shape `(..., nat, nat)`.
        """
        # The default einsum path is fastest if the large tensors comes first.
        # (..., n1, n2, r1, r2) * (..., n1, r1) * (..., n2, r2) -> (..., n1, n2)
        return einsum(
            "...ijab,...ia,...jb->...ij",
            *(self.rc6, gw, gw),
            optimize=[(0, 1), (0, 1)],
        )

    def get_weighted_pols(self, gw: Tensor) -> Tensor:
        """
        Calculate the weighted polarizabilities for each atom and frequency.

        Parameters
        ----------
        gw : Tensor
            Weights for the atomic reference systems of shape
            ``(..., nat, nref)``.

        Returns
        -------
        Tensor
            Weighted polarizabilities of shape ``(..., nat, 23)``.
        """
        a = self._get_alpha()
        return einsum("...nr,...nrw->...nw", gw, a)



# This file is part of tad-dftd4.
#
# SPDX-Identifier: Apache-2.0
# Copyright (C) 2024 Grimme Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Damping Parameters
==================

Read damping parameters from toml file. The TOML file is coped from the DFT-D4
Fortran GitHub repository.
(https://github.com/dftd4/dftd4/blob/main/assets/parameters.toml)
"""
from __future__ import annotations

from enum import Enum
from typing import Optional, Type

from pydantic import BaseModel, Field
from tad_mctc.typing import Tensor, TypedDict
from typing_extensions import NotRequired

__all__ = [
    "Param",
    "DispersionMethod",
    "Variant",
    "BaseParams",
    "RationalParams",
    "ZeroParams",
    "MZeroParams",
    "OptimisedPowerParams",
    "PARAM_MODELS",
]


class Param(TypedDict, total=False):
    """Type annotation for dispersion parameters."""

    a1: NotRequired[Tensor | float]
    """Scaling for the C8 / C6 ratio in the critical radius (0.4)."""

    a2: NotRequired[Tensor | float]
    """Offset parameter for the critical radius (5.0)."""

    s6: NotRequired[Tensor | float]
    """Scaling of dipole-dipole term (1.0 to retain correct limit)."""

    rs6: NotRequired[Tensor | float]
    """Scaling of dipole-dipole term for (modified) zero damping (1.0)."""

    s8: NotRequired[Tensor | float]
    """Scaling of dipole-quadrupole term (1.0)."""

    rs8: NotRequired[Tensor | float]
    """Scaling of dipole-quadrupole term for (modified) zero damping (1.0)."""

    s9: NotRequired[Tensor | float]
    """Scaling of three-body term (1.0)."""

    rs9: NotRequired[Tensor | float]
    """Scaling for van-der-Waals radii in damping function (4.0/3.0)."""

    s10: NotRequired[Tensor | float]
    """Scaling of quadrupole-quadrupole term (0.0)."""

    alp: NotRequired[Tensor | float]
    """Exponent of zero damping function (16.0)."""

    bet: NotRequired[Tensor | float]
    """Exponent of mzero or optimized-power damping function (0.0)."""

    doi: NotRequired[str]
    """DOI of the reference paper."""


class DispersionMethod(str, Enum):
    d3 = "d3"
    d4 = "d4"
    d5 = "d5"


class Variant(str, Enum):
    bj = "bj"
    bj_eeq_atm = "bj-eeq-atm"  # D4 default
    zero = "zero"
    mzero = "mzero"
    optimizedpower = "optimizedpower"


# Parameter data classes


class BaseParams(BaseModel):
    """Fields shared by *all* damping families."""

    s6: float = Field(1.0)
    s8: float = Field(1.0)
    s10: float = Field(0.0)
    doi: Optional[str] = Field(None, description="Optional reference")


class RationalParams(BaseParams):
    a1: float
    a2: float


class ZeroParams(BaseParams):
    alp: float
    rs6: float
    rs8: float
    rs9: float = Field(0.0, description="Radius scaling for MB term")
    bet: float = Field(0.0, description="exponent shift β (optional for zero)")


class MZeroParams(ZeroParams):
    bet: float = Field(
        ...,  # ellipsis indicates this field is required
        description="exponent shift β (required for m-zero damping)",
    )


class OptimisedPowerParams(BaseParams):
    a1: float
    a2: float
    bet: float


PARAM_MODELS: dict[Variant, Type[BaseParams]] = {
    Variant.bj: RationalParams,
    Variant.bj_eeq_atm: RationalParams,  # same functional form
    Variant.zero: ZeroParams,
    Variant.mzero: MZeroParams,
    Variant.optimizedpower: OptimisedPowerParams,
}
"""Registry of parameter models for each damping variant."""



# This file is part of tad-dftd4.
#
# SPDX-Identifier: Apache-2.0
# Copyright (C) 2024 Grimme Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Dispersion: 2-body terms
==========================

Implementation of the 2-body dispersion terms.
"""
from __future__ import annotations

import torch
from tad_mctc import storch
from tad_mctc.batch import real_pairs
from tad_mctc.typing import DD, Tensor

from .. import defaults
from ..cutoff import Cutoff
from ..damping import Damping, Param, RationalDamping
from ..model import ModelInst
from .base import DispTerm


class TwoBodyTerm(DispTerm):
    """
    Base class for two-body dispersion terms.

    Parameters
    ----------
    damping_fn : Damping, optional
        Damping function to evaluate distance-dependent contributions.
        Defaults to :class:`.RationalDamping`.
    charge_dependent : bool, optional
        Whether the term is charge-dependent. Defaults to ``True``.
    """

    def __init__(
        self,
        *,
        damping_fn: Damping = RationalDamping(),
        charge_dependent: bool = True,
    ):
        super().__init__(damping_fn, charge_dependent)
        self.damping_fn = damping_fn

    def calculate(
        self,
        numbers: Tensor,
        positions: Tensor,
        param: Param,
        cn: Tensor,
        model: ModelInst,
        q: Tensor | None,
        r4r2: Tensor,
        rvdw: Tensor,
        cutoff: Cutoff,
    ) -> Tensor:
        weights = model.weight_references(
            cn, q if self.charge_dependent else None
        )
        c6 = model.get_atomic_c6(weights)

        return dispersion2(
            numbers,
            positions,
            param,
            c6,
            r4r2,
            rvdw,
            self.damping_fn,
            cutoff.disp2,
        )


def dispersion2(
    numbers: Tensor,
    positions: Tensor,
    param: Param,
    c6: Tensor,
    r4r2: Tensor,
    rvdw: Tensor | None = None,
    damping_function: Damping = RationalDamping(),
    cutoff: Tensor | None = None,
    as_matrix: bool = False,
) -> Tensor:
    """
    Calculate dispersion energy between pairs of atoms.

    Parameters
    ----------
    numbers : Tensor
        Atomic numbers for all atoms in the system of shape ``(..., nat)``.
    positions : Tensor
        Cartesian coordinates of all atoms (shape: ``(..., nat, 3)``).
    param : Param
        DFT-D3 damping parameters.
    c6 : Tensor
        Atomic C6 dispersion coefficients.
    r4r2 : Tensor
        r⁴ over r² expectation values of the atoms in the system.
    rvdw : Tensor
        Pair-wise van der Waals radii of the atoms in the system.
    damping_function : Damping, optional
        Damping function evaluate distance dependent contributions.
        Additional arguments are passed through to the function.
        Defaults to :class:`.RationalDamping`.
    cutoff : Tensor | None, optional
        Real-space cutoff for two-body dispersion. Defaults to `None`, which
        will be evaluated to `defaults.D4_DISP2_CUTOFF`.
    as_matrix : bool, optional
        Return dispersion energy as a matrix. If you sum up the dispersion
        energy from the matrix, do not forget the factor `0.5` that fixes the
        double counting. Defaults to `False`.

    Returns
    -------
    Tensor
        Atom-resolved two-body dispersion energy.
    """
    dd: DD = {"device": positions.device, "dtype": positions.dtype}
    zero = torch.tensor(0.0, **dd)

    if cutoff is None:
        cutoff = torch.tensor(defaults.D4_DISP2_CUTOFF, **dd)

    mask = real_pairs(numbers, mask_diagonal=True)
    distances = torch.where(
        mask,
        storch.cdist(positions, positions, p=2),
        torch.tensor(torch.finfo(positions.dtype).eps, **dd),
    )

    qq = 3 * r4r2.unsqueeze(-1) * r4r2.unsqueeze(-2)
    c8 = c6 * qq

    if damping_function.radius_type == "rvdw":
        if rvdw is None:
            raise ValueError("`rvdw` must be provided for `rvdw` radius type.")
        radii = rvdw
    elif damping_function.radius_type == "r4r2":
        radii = qq
    else:
        raise ValueError(f"Unknown radius type: {damping_function.radius_type}")

    t6 = torch.where(
        mask * (distances <= cutoff),
        damping_function(distances, radii, 6, **param),
        zero,
    )
    t8 = torch.where(
        mask * (distances <= cutoff),
        damping_function(distances, radii, 8, **param),
        zero,
    )

    if as_matrix is True:
        e6 = c6 * t6
        e8 = c8 * t8
    else:
        e6 = torch.sum(c6 * t6, dim=-1)
        e8 = torch.sum(c8 * t8, dim=-1)

    s6 = param.get("s6", torch.tensor(defaults.S6, **dd))
    s8 = param.get("s8", torch.tensor(defaults.S8, **dd))

    edisp = s6 * e6 + s8 * e8

    # With `if "s10" in param and param["s10"] != 0.0`, the gradcheck tests fail
    # if s10 is exactly 0 (other values are fine).
    if "s10" in param:
        c10 = c6 * torch.pow(qq, 2) * 49.0 / 40.0
        t10 = torch.where(
            mask * (distances <= cutoff),
            damping_function(distances, radii, 10, **param),
            zero,
        )

        if as_matrix is True:
            e10 = c10 * t10
        else:
            e10 = torch.sum(c10 * t10, dim=-1)

        edisp += param["s10"] * e10

    if as_matrix is True:
        return -edisp
    return -0.5 * edisp



# This file is part of tad-dftd4.
#
# SPDX-Identifier: Apache-2.0
# Copyright (C) 2024 Grimme Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Model: Base
===========

This module contains the definition of the base dispersion model for the
evaluation of C6 coefficients.

Upon instantiation, the reference polarizabilities are calculated for the
all atoms of the molecule(s) and stored in the model class.


Example
-------
>>> import torch
>>> import tad_dftd4 as d4
>>>
>>> numbers = torch.tensor([14, 1, 1, 1, 1]) # SiH4
>>> model = d4.D4Model(numbers)
>>>
>>> # calculate Gaussian weights, optionally pass CN and partial charges
>>> gw = model.weight_references()
>>> c6 = model.get_atomic_c6(gw)
"""
from __future__ import annotations

from abc import abstractmethod

import torch
from tad_mctc import storch
from tad_mctc.convert import any_to_tensor
from tad_mctc.math import einsum
from tad_mctc.typing import Literal, Tensor, TensorLike, overload

from .. import data
from ..utils import trapzd

__all__ = ["BaseModel", "WF_DEFAULT"]


GA_DEFAULT = 3.0
GC_DEFAULT = 2.0
WF_DEFAULT = 6.0


class BaseModel(TensorLike):
    """
    The D4 dispersion model.
    """

    numbers: Tensor
    """Atomic numbers of all atoms in the system."""

    ga: float
    """
    Maximum charge scaling height for partial charge extrapolation.

    :default: :data:`.GA_DEFAULT`
    """

    gc: float
    """
    Charge scaling steepness for partial charge extrapolation.

    :default: :data:`.GC_DEFAULT`
    """

    wf: Tensor
    """
    Weighting factor for coordination number interpolation.

    :default: ``None`` (model-dependent, set upon instantiation)
    """

    ref_charges: Literal["eeq", "gfn2"]
    """
    Reference charges to use for the model.

    :default: ``"eeq"``
    """

    rc6: Tensor
    """
    Reference C6 coefficients of all atoms.

    :default: ``None`` (calculated upon instantiation)
    """

    __slots__ = ("numbers", "ga", "gc", "wf", "ref_charges", "rc6")

    def __init__(
        self,
        numbers: Tensor,
        ga: float = GA_DEFAULT,
        gc: float = GC_DEFAULT,
        wf: Tensor | float | None = None,
        ref_charges: Literal["eeq", "gfn2"] = "eeq",
        rc6: Tensor | None = None,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> None:
        """
        Instantiate `D4Model`.

        Parameters
        ----------
        numbers : Tensor
            Atomic numbers of all atoms in the system.
        ga : float, optional
            Maximum charge scaling height for partial charge extrapolation.
            Defaults to `GA_DEFAULT`.
        gc : float, optional
            Charge scaling steepness for partial charge extrapolation.
            Defaults to `GC_DEFAULT`.
        wf : float, optional
            Weighting factor for coordination number interpolation.
            Defaults to `WF_DEFAULT`.
        ref_charges : Literal["eeq", "gfn2"], optional
            Reference charges to use for the model. Defaults to `"eeq"`.
        rc6 : Tensor | None, optional
            Reference C6 coefficients of all atoms. Defaults to `None`.
        device : torch.device | None, optional
            Pytorch device for calculations. Defaults to `None`.
        dtype : torch.dtype | None, optional
            Pytorch dtype for calculations. Defaults to `None`.
        """
        super().__init__(device, dtype)
        self.numbers = numbers

        self.ga = ga
        self.gc = gc
        self.ref_charges = ref_charges

        self.wf = self._get_wf() if wf is None else any_to_tensor(wf, **self.dd)
        self.rc6 = self._get_refc6() if rc6 is None else rc6

    ####################
    # Abstract methods #
    ####################

    @abstractmethod
    def _get_wf(self) -> Tensor:
        """
        Get the weighting factor for the Gaussian weights.

        Returns
        -------
        Tensor
            Weighting factor for the Gaussian weights.
        """

    @abstractmethod
    def get_atomic_c6(self, gw: Tensor) -> Tensor:
        """
        Calculate atomic C6 dispersion coefficients.

        Parameters
        ----------
        gw : Tensor
            Weights for the atomic reference systems of shape
            `(..., nat, nref)`.

        Returns
        -------
        Tensor
            C6 coefficients for all atom pairs of shape `(..., nat, nat)`.
        """

    @overload
    @abstractmethod
    def weight_references(
        self,
        cn: Tensor | None = None,
        q: Tensor | None = None,
        *,
        with_dgwdq: Literal[False] = ...,
        with_dgwdcn: Literal[False] = ...,
    ) -> Tensor: ...

    @overload
    @abstractmethod
    def weight_references(
        self,
        cn: Tensor | None = None,
        q: Tensor | None = None,
        *,
        with_dgwdq: Literal[True],
        with_dgwdcn: Literal[False] = ...,
    ) -> tuple[Tensor, Tensor]: ...

    @overload
    @abstractmethod
    def weight_references(
        self,
        cn: Tensor | None = None,
        q: Tensor | None = None,
        *,
        with_dgwdq: Literal[False] = ...,
        with_dgwdcn: Literal[True],
    ) -> tuple[Tensor, Tensor]: ...

    @overload
    @abstractmethod
    def weight_references(
        self,
        cn: Tensor | None = None,
        q: Tensor | None = None,
        *,
        with_dgwdq: Literal[True],
        with_dgwdcn: Literal[True],
    ) -> tuple[Tensor, Tensor, Tensor]: ...

    @abstractmethod
    def weight_references(
        self,
        cn: Tensor | None = None,
        q: Tensor | None = None,
        *,
        with_dgwdq: bool = False,
        with_dgwdcn: bool = False,
    ) -> Tensor | tuple[Tensor, Tensor] | tuple[Tensor, Tensor, Tensor]:
        """
        Calculate the weights of the reference system
        (shape: ``(..., nat, nref)``).

        Parameters
        ----------
        cn : Tensor | None, optional
            Coordination number of every atom. Defaults to `None` (0).
        q : Tensor | None, optional
            Partial charge of every atom. Defaults to `None` (0).
        with_dgwdq : bool, optional
            Whether to also calculate the derivative of the weights with
            respect to the partial charges. Defaults to `False`.
        with_dgwdcn : bool, optional
            Whether to also calculate the derivative of the weights with
            respect to the coordination numbers. Defaults to `False`.

        Returns
        -------
        Tensor | tuple[Tensor, Tensor] | tuple[Tensor, Tensor, Tensor]
            Weights for the atomic reference systems (shape:
            ``(..., nat, ref)``). If ``with_dgwdq`` is ``True``, also returns
            the derivative of the weights with respect to the partial charges.
            If ``with_dgwdcn`` is ``True``, also returns the derivative of the
            weights with respect to the coordination numbers.
        """

    @abstractmethod
    def get_weighted_pols(self, gw: Tensor) -> Tensor:
        """
        Calculate the weighted polarizabilities for each atom and frequency.

        Parameters
        ----------
        gw : Tensor
            Weights for the atomic reference systems of shape
            ``(..., nat, nref)``.

        Returns
        -------
        Tensor
            Weighted polarizabilities of shape ``(..., nat, 23)``.
        """

    ##################
    # Public methods #
    ##################

    def get_polarizabilities(self, weights: Tensor) -> Tensor:
        """
        Calculate static polarizabilities for all atoms.

        Parameters
        ----------
        weights : Tensor
            Weights for the atomic reference systems of shape
            ``(..., nat, nref)``.

        Returns
        -------
        Tensor
            Polarizabilities of shape ``(..., nat)``.
        """
        # (..., n, r) * (..., n, r) -> (..., n)
        return einsum("...nr,...nr->...n", weights, self._get_alpha()[..., 0])

    ###################
    # Private methods #
    ###################

    def _zeta(self, gam: Tensor, qref: Tensor, qmod: Tensor) -> Tensor:
        """
        Charge scaling function.

        Parameters
        ----------
        gam : Tensor
            Chemical hardness.
        qref : Tensor
            Reference charges.
        qmod : Tensor
            Modified charges.

        Returns
        -------
        Tensor
            Scaled charges.
        """
        eps = torch.tensor(torch.finfo(self.dtype).eps, **self.dd)
        ga = torch.tensor(self.ga, **self.dd)

        scale = torch.exp(gam * (1.0 - qref / (qmod - eps)))

        return torch.where(
            qmod > 0.0,
            torch.exp(ga * (1.0 - scale)),
            torch.exp(ga),
        )

    def _dzeta(self, gam: Tensor, qref: Tensor, qmod: Tensor) -> Tensor:
        """
        Derivative of charge scaling function with respect to `qmod`.

        Parameters
        ----------
        gam : Tensor
            Chemical hardness.
        qref : Tensor
            Reference charges.
        qmod : Tensor
            Modified charges.

        Returns
        -------
        Tensor
            Derivative of charges.
        """
        eps = torch.tensor(torch.finfo(self.dtype).eps, **self.dd)
        ga = torch.tensor(self.ga, **self.dd)

        scale = torch.exp(gam * (1.0 - qref / (qmod - eps)))
        zeta = torch.exp(ga * (1.0 - scale))

        return torch.where(
            qmod > 0.0,
            -ga * gam * scale * zeta * storch.divide(qref, qmod**2),
            torch.tensor(0.0, **self.dd),
        )

    def _get_alpha(self) -> Tensor:
        """
        Calculate reference polarizabilities.

        Returns
        -------
        Tensor
            Reference polarizabilities of shape `(..., nat, ref, 23)`.
        """
        # pylint: disable=import-outside-toplevel
        from ..reference import d4 as reference

        zero = torch.tensor(0.0, **self.dd)

        refsys = reference.refsys.to(self.device)[self.numbers]
        refascale = reference.refascale.to(**self.dd)[self.numbers]
        refalpha = reference.refalpha.to(**self.dd)[self.numbers]
        refscount = reference.refscount.to(**self.dd)[self.numbers]
        secscale = reference.secscale.to(**self.dd)
        secalpha = reference.secalpha.to(**self.dd)

        if self.ref_charges == "eeq":
            # pylint: disable=import-outside-toplevel
            from ..reference.d4.charge_eeq import clsh as _refsq

            refsq = _refsq.to(**self.dd)[self.numbers]
        elif self.ref_charges == "gfn2":
            # pylint: disable=import-outside-toplevel
            from ..reference.d4.charge_gfn2 import refh as _refsq

            refsq = _refsq.to(**self.dd)[self.numbers]
        else:
            raise ValueError(f"Unknown reference charges: {self.ref_charges}")

        mask = refsys > 0

        zeff = data.ZEFF(self.device)[refsys]
        gam = data.GAM(**self.dd)[refsys] * self.gc

        # charge scaling
        zeta = torch.where(
            mask,
            self._zeta(gam, zeff, refsq + zeff),
            zero,
        )

        aiw = secscale[refsys] * secalpha[refsys] * zeta.unsqueeze(-1)
        h = refalpha - refscount.unsqueeze(-1) * aiw
        alpha = refascale.unsqueeze(-1) * h

        # (..., n, r, 23)
        return torch.where(alpha > 0.0, alpha, zero)

    def _get_refc6(self) -> Tensor:
        """
        Calculate reference C6 dispersion coefficients. The reference C6
        coefficients are not weighted by the Gaussian weights yet.

        Returns
        -------
        Tensor
            Reference C6 coefficients of shape ``(..., nat, nat, nref, nref)``.
        """
        # (..., n, r, 23) -> (..., n, n, r, r)
        return trapzd(self._get_alpha())

    ############
    # Printing #
    ############

    def __str__(self) -> str:  # pragma: no cover
        """Return a string representation of the model."""
        return (
            f"{self.__class__.__name__}(\n"
            f"  ga={self.ga},\n"
            f"  gc={self.gc},\n"
            f"  wf={self.wf},\n"
            f"  ref_charges={self.ref_charges},\n"
            f"  rc6={self.rc6.shape},\n"
            f")"
        )

    def __repr__(self) -> str:  # pragma: no cover
        """Return a string representation of the model."""
        return str(self)



# This file is part of tad-dftd4.
#
# SPDX-Identifier: Apache-2.0
# Copyright (C) 2024 Grimme Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Dispersion Methods: D3
======================

DFT-D3 dispersion method implementation.
"""
from __future__ import annotations

from ..damping import RationalDamping, ZeroDamping
from .base import Disp
from .threebody import ATM, C9ApproxMixin, RadiiVDWMixin
from .twobody import TwoBodyTerm

__all__ = ["DispD3BJ", "DispD3Zero"]


class _DispD3(Disp):
    """Standard DFT-D3 dispersion method."""

    ALLOWED_MODELS = ("d3",)
    """Allowed DFT-D models for the calculation."""


##############################################################################


class D3ATM(C9ApproxMixin, RadiiVDWMixin, ATM):
    """D3 ATM: approximate C9 + pair-wise VDW radii."""


class DispD3BJ(_DispD3):
    """Standard DFT-D3 dispersion method."""

    TERMS = [
        (
            TwoBodyTerm,
            {"damping_fn": RationalDamping(), "charge_dependent": False},
        ),
        (
            D3ATM,
            {"damping_fn": ZeroDamping(), "charge_dependent": False},
        ),
    ]
    """List of dispersion terms to be registered in the constructor."""


##############################################################################


class DispD3Zero(_DispD3):
    """Zero-damping DFT-D3 dispersion method."""

    TERMS = [
        (
            TwoBodyTerm,
            {"damping_fn": ZeroDamping(), "charge_dependent": False},
        ),
        (
            D3ATM,
            {"damping_fn": ZeroDamping(), "charge_dependent": False},
        ),
    ]
    """List of dispersion terms to be registered in the constructor."""



# This file is part of tad-dftd4.
#
# SPDX-Identifier: Apache-2.0
# Copyright (C) 2024 Grimme Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Dispersion: 3-body terms
========================

Implementation of the 3-body Axilrod-Teller-Muto dispersion terms.

.. math::

    E_\text{disp}^{(3), \text{ATM}} &=
    \sum_\text{ABC} E^{\text{ABC}} f_\text{damp}\left(\overline{R}_\text{ABC}\right) \\
    E^{\text{ABC}} &=
    \dfrac{C^{\text{ABC}}_9
    \left(3 \cos\theta_\text{A} \cos\theta_\text{B} \cos\theta_\text{C} + 1 \right)}
    {\left(r_\text{AB} r_\text{BC} r_\text{AC} \right)^3} \\
    f_\text{damp} &=
    \dfrac{1}{1+ 6 \left(\overline{R}_\text{ABC}\right)^{-16}}
"""
from __future__ import annotations

from abc import ABC, abstractmethod

import torch
from tad_mctc import storch
from tad_mctc.batch import real_pairs, real_triples
from tad_mctc.convert import any_to_tensor
from tad_mctc.typing import DD, Tensor

from .. import defaults
from ..cutoff import Cutoff
from ..damping import Damping, Param, ZeroDamping
from ..model import ModelInst
from .base import DispTerm

__all__ = ["ATM", "get_atm_dispersion"]


def get_atm_dispersion(
    numbers: Tensor,
    positions: Tensor,
    c9: Tensor,
    radii: Tensor,
    cutoff: Tensor,
    damping_function: Damping = ZeroDamping(),
    s9: Tensor | float | int = defaults.S9,
    alp: Tensor | float | int = defaults.ALP,
) -> Tensor:
    """
    Axilrod-Teller-Muto dispersion term.

    Parameters
    ----------
    numbers : Tensor
        Atomic numbers for all atoms in the system of shape ``(..., nat)``.
    positions : Tensor
        Cartesian coordinates of all atoms (shape: ``(..., nat, 3)``).
    c9 : Tensor
        Atomic C9 dispersion coefficients.
    radii : Tensor
        Pairwise critical radii for all atom pairs (shape: ``(..., nat, nat)``).
    cutoff : Tensor
        Real-space cutoff.
    damping_function : Damping, optional
        Damping function to use. Defaults to :func:`zero_damping`.
    s9 : Tensor, optional
        Scaling for dispersion coefficients. Defaults to ``1.0``.
    alp : Tensor, optional
        Exponent of zero damping function. Defaults to ``14.0``.

    Returns
    -------
    Tensor
        Atom-resolved ATM dispersion energy.
    """
    dd: DD = {"device": positions.device, "dtype": positions.dtype}

    s9 = any_to_tensor(s9, **dd)
    alp = any_to_tensor(alp, **dd)

    cutoff2 = cutoff * cutoff

    mask_pairs = real_pairs(numbers, mask_diagonal=True)
    mask_triples = real_triples(numbers, mask_diagonal=True, mask_self=True)

    # filler values for masks
    eps = torch.tensor(torch.finfo(positions.dtype).eps, **dd)
    zero = torch.tensor(0.0, **dd)
    one = torch.tensor(1.0, **dd)

    r0ij = radii.unsqueeze(-1)
    r0ik = radii.unsqueeze(-2)
    r0jk = radii.unsqueeze(-3)
    r0 = r0ij * r0ik * r0jk

    # actually faster than other alternatives
    # very slow: (pos.unsqueeze(-2) - pos.unsqueeze(-3)).pow(2).sum(-1)
    distances = torch.pow(
        torch.where(
            mask_pairs,
            storch.cdist(positions, positions, p=2),
            eps,
        ),
        2.0,
    )

    r2ij = distances.unsqueeze(-1)
    r2ik = distances.unsqueeze(-2)
    r2jk = distances.unsqueeze(-3)
    r2 = r2ij * r2ik * r2jk
    r1 = torch.sqrt(r2)
    # add epsilon to avoid zero division later
    r3 = torch.where(mask_triples, r1 * r2, eps)
    r5 = torch.where(mask_triples, r2 * r3, eps)

    # to fix the previous mask, we mask again (not strictly necessary because
    # `ang` is also masked and we later multiply with `ang`)
    fdamp = torch.where(
        mask_triples,
        damping_function(
            r0,
            # dividing by tiny numbers leads to huge numbers, which result in
            # NaN's upon exponentiation in the subsequent step
            torch.where(mask_triples, r1, one),
            order=9,
            alp=alp / 3.0,
            only_damping=True,
        ),
        zero,
    )

    s = torch.where(
        mask_triples,
        (r2ij + r2jk - r2ik) * (r2ij - r2jk + r2ik) * (-r2ij + r2jk + r2ik),
        zero,
    )

    ang = torch.where(
        mask_triples
        * (r2ij <= cutoff2)
        * (r2jk <= cutoff2)
        * (r2jk <= cutoff2),
        0.375 * s / r5 + 1.0 / r3,
        torch.tensor(0.0, **dd),
    )

    energy = ang * fdamp * s9 * c9
    return torch.sum(energy, dim=(-2, -1)) / 6.0


class ThreeBodyTerm(DispTerm):
    """Base class for three-body dispersion terms."""

    def __init__(
        self,
        *,
        damping_fn: Damping = ZeroDamping(),
        charge_dependent: bool = False,
    ):
        super().__init__(damping_fn, charge_dependent)


class ATM(ThreeBodyTerm, ABC):
    r"""
    D4's Axilrod-Teller-Muto dispersion term.
    - C9 coefficients are approximated from C6 coefficients
    - Becke--Johnson cutoff radii (`a1 * sqrt(3.0 * r4r2) + a2`)
    - zero damping

    .. math::

        E_\text{disp}^{(3), \text{ATM}} &=
        \sum_\text{ABC} E^{\text{ABC}} f_\text{damp}\left(\overline{R}_\text{ABC}\right) \\
        E^{\text{ABC}} &=
        \dfrac{C^{\text{ABC}}_9
        \left(3 \cos\theta_\text{A} \cos\theta_\text{B} \cos\theta_\text{C} + 1 \right)}
        {\left(r_\text{AB} r_\text{BC} r_\text{AC} \right)^3} \\
        f_\text{damp} &=
        \dfrac{1}{1+ 6 \left(\overline{R}_\text{ABC}\right)^{-16}}
    """

    @abstractmethod
    def get_c9(self, model: ModelInst, cn: Tensor, q: Tensor | None) -> Tensor:
        """Approximate or exact C9 coefficients."""

    @abstractmethod
    def get_radii(
        self,
        param: Param,
        r4r2: Tensor,
        rvdw: Tensor,
    ) -> Tensor:
        """Compute critical radii used in damping function."""

    def calculate(
        self,
        numbers: Tensor,
        positions: Tensor,
        param: Param,
        cn: Tensor,
        model: ModelInst,
        q: Tensor | None,
        r4r2: Tensor,
        rvdw: Tensor,
        cutoff: Cutoff,
    ):
        # ATM-specific C9 coefficients and radii
        c9 = self.get_c9(model, cn, q)
        radii = self.get_radii(param, r4r2, rvdw)

        return get_atm_dispersion(
            numbers,
            positions,
            c9,
            radii,
            damping_function=self.damping_fn,
            cutoff=cutoff.disp3,
            s9=param.get("s9", torch.tensor(defaults.S9, **self.dd)),
            alp=param.get("alp", torch.tensor(defaults.ALP, **self.dd)),
        )


# Radii mixins


class RadiiBJMixin:
    """Becke--Johnson critical radii: a1 * sqrt(3 * r4r2_i * r4r2_j) + a2."""

    def get_radii(
        self,
        param: Param,
        r4r2: Tensor,
        rvdw: Tensor,
    ) -> Tensor:
        dd = self.dd  # type: ignore

        a1 = param.get("a1", torch.tensor(defaults.A1, **dd))
        a2 = param.get("a2", torch.tensor(defaults.A2, **dd))
        return (
            a1 * storch.sqrt(3.0 * r4r2.unsqueeze(-1) * r4r2.unsqueeze(-2)) + a2
        )


class RadiiVDWMixin:
    """Scaled pair-wise VDW radii: rs9 * rvdw."""

    def get_radii(
        self,
        param: Param,
        r4r2: Tensor,
        rvdw: Tensor,
    ) -> Tensor:
        dd = self.dd  # type: ignore
        rs9 = param.get("rs9", torch.tensor(defaults.RS9, **dd))
        return rs9 * rvdw


# C9 mixins


class C9ExactMixin:
    """Exact C9 via Casimir–Polder integration of polarizabilities."""

    charge_dependent: bool
    """Whether the C9 coefficients depend on atomic charges."""

    def get_c9(self, model: ModelInst, cn: Tensor, q: Tensor | None) -> Tensor:
        r"""
        Approximate C9 coefficients from C6 coefficients.

        .. math::

            C_9 = \sqrt{|C_{6}^{AB} C_{6}^{AC} C_{6}^{BC}|}
        """
        # pylint: disable=import-outside-toplevel
        from ..utils import trapzd_atm

        weights = model.weight_references(
            cn, q if self.charge_dependent else None
        )
        aiw = model.get_weighted_pols(weights)

        # C9_ABC = integral (aiw_A * aiw_B * aiw_C)
        aiwi = aiw.unsqueeze(-2).unsqueeze(-2)
        aiwj = aiw.unsqueeze(-3).unsqueeze(-2)
        aiwk = aiw.unsqueeze(-3).unsqueeze(-3)
        return trapzd_atm(aiwi * aiwj * aiwk)


class C9ApproxMixin:
    """Approximate C9 coefficients from C6 coefficients."""

    charge_dependent: bool
    """Whether the C9 coefficients depend on atomic charges."""

    def get_c9(self, model: ModelInst, cn: Tensor, q: Tensor | None) -> Tensor:

        weights = model.weight_references(
            cn, q if self.charge_dependent else None
        )
        c6 = model.get_atomic_c6(weights)

        # C9_ABC = sqrt(|C6_AB * C6_AC * C6_BC|)
        return storch.sqrt(
            torch.abs(c6.unsqueeze(-1) * c6.unsqueeze(-2) * c6.unsqueeze(-3)),
        )



# This file is part of tad-dftd4.
#
# SPDX-Identifier: Apache-2.0
# Copyright (C) 2024 Grimme Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Defaults
========

This module defines the default values for all parameters within `tad_dftd4`.
"""

# DFT-D4

D4_CN_CUTOFF = 30.0
"""Coordination number cutoff (30.0)."""

D4_CN_EEQ_CUTOFF = 25.0
"""Coordination number cutoff within EEQ (25.0)."""

D4_CN_EEQ_MAX = 8.0
"""Maximum coordination number (8.0)."""

D4_DISP2_CUTOFF = 60.0
"""Two-body interaction cutoff (60.0)."""

D4_DISP3_CUTOFF = 40.0
"""Three-body interaction cutoff (40.0)."""

D4_KCN = 7.5
"""Steepness of counting function (7.5)."""

D4_K4 = 4.10451
"""Parameter for electronegativity scaling."""

D4_K5 = 19.08857
"""Parameter for electronegativity scaling."""

D4_K6 = 2 * 11.28174**2  # 254.56
"""Parameter for electronegativity scaling."""

# DFT-D4 damping parameters

A1 = 0.4
"""Scaling for the C8 / C6 ratio in the critical radius (0.4)."""

A2 = 5.0
"""Offset parameter for the critical radius (5.0)."""

RS6 = 1.0
"""
Default scaling of dipole-dipole term for (modified) zero damping (1.0).
"""

S6 = 1.0
"""Default scaling of dipole-dipole term (1.0 to retain correct limit)."""

RS8 = 1.0
"""
Default scaling of dipole-quadrupole term for (modified) zero damping (1.0).
"""

S8 = 1.0
"""Default scaling of dipole-quadrupole term (1.0)."""

S9 = 1.0
"""Default scaling of three-body term (1.0)."""

S10 = 0.0
"""Default scaling of quadrupole-quadrupole term (0.0)."""

RS9 = 1.0  # 4.0 / 3.0
"""
Scaling for van-der-Waals radii in damping function.
- D3: 4.0 / 3.0
- D4: 1.0
"""

ALP = 16.0
"""Exponent of zero damping function (16.0)."""

BET = 0.0
"""Exponent of mzero or optimized-power damping function (0.0)."""



# This file is part of tad-dftd4.
#
# SPDX-Identifier: Apache-2.0
# Copyright (C) 2024 Grimme Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Torch autodiff DFT-D4
=====================

Implementation of the DFT-D4 dispersion model in PyTorch.
This module allows to process a single structure or a batch of structures for
the calculation of atom-resolved dispersion energies.

.. note::

   This project is still in early development and the API is subject to change.
   Contributions are welcome, please checkout our
   `contributing guidelines <https://github.com/dftd4/tad-dftd4/blob/main/CONTRIBUTING.md>`_.

Example
-------
>>> import torch
>>> import tad_dftd4 as d4
>>> import tad_mctc as mctc
>>>
>>> # S22 system 4: formamide dimer
>>> numbers = mctc.batch.pack((
...     mctc.convert.symbol_to_number("C C N N H H H H H H O O".split()),
...     mctc.convert.symbol_to_number("C O N H H H".split()),
... ))
>>>
>>> # coordinates in Bohr
>>> positions = mctc.batch.pack((
...     torch.tensor([
...         [-3.81469488143921, +0.09993441402912, 0.00000000000000],
...         [+3.81469488143921, -0.09993441402912, 0.00000000000000],
...         [-2.66030049324036, -2.15898251533508, 0.00000000000000],
...         [+2.66030049324036, +2.15898251533508, 0.00000000000000],
...         [-0.73178529739380, -2.28237795829773, 0.00000000000000],
...         [-5.89039325714111, -0.02589114569128, 0.00000000000000],
...         [-3.71254944801331, -3.73605775833130, 0.00000000000000],
...         [+3.71254944801331, +3.73605775833130, 0.00000000000000],
...         [+0.73178529739380, +2.28237795829773, 0.00000000000000],
...         [+5.89039325714111, +0.02589114569128, 0.00000000000000],
...         [-2.74426102638245, +2.16115570068359, 0.00000000000000],
...         [+2.74426102638245, -2.16115570068359, 0.00000000000000],
...     ]),
...     torch.tensor([
...         [-0.55569743203406, +1.09030425468557, 0.00000000000000],
...         [+0.51473634678469, +3.15152550263611, 0.00000000000000],
...         [+0.59869690244446, -1.16861263789477, 0.00000000000000],
...         [-0.45355203669134, -2.74568780438064, 0.00000000000000],
...         [+2.52721209544999, -1.29200800956867, 0.00000000000000],
...         [-2.63139587595376, +0.96447869452240, 0.00000000000000],
...     ]),
... ))
>>>
>>> # total charge of both systems
>>> charge = torch.tensor([0.0, 0.0])
>>>
>>> # TPSSh-D4-ATM parameters
>>> param = {
...     "s6": torch.tensor(1.0),
...     "s8": torch.tensor(1.85897750),
...     "s9": torch.tensor(1.0),
...     "a1": torch.tensor(0.44286966),
...     "a2": torch.tensor(4.60230534),
... }
>>>
>>> # calculate dispersion energy in Hartree
>>> energy = torch.sum(d4.dftd4(numbers, positions, charge, param), -1)
>>> torch.set_printoptions(precision=10)
>>> print(energy)
tensor([-0.0088341432, -0.0027013607])
>>> print(energy[0] - 2*energy[1])
tensor(-0.0034314217)
"""
import torch

from . import cutoff, damping, data, dispersion, model
from .__version__ import __version__
from .cutoff import Cutoff
from .damping import Param, get_params
from .disp import dftd4, get_properties
from .model import D4Model, D4SModel

__all__ = [
    "__version__",
    "cutoff",
    "Cutoff",
    "damping",
    "data",
    "dispersion",
    "dftd4",
    "disp",
    "get_params",
    "get_properties",
    "model",
    "D4Model",
    "D4SModel",
    "Param",
]



# This file is part of tad-dftd4.
#
# SPDX-Identifier: Apache-2.0
# Copyright (C) 2024 Grimme Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Damping Parameters
==================

Read damping parameters from toml file. The TOML file is coped from the
Fortran GitHub repository.
- https://github.com/dftd3/s-dftd3/blob/main/assets/parameters.toml
- https://github.com/dftd4/dftd4/blob/main/assets/parameters.toml
"""
from __future__ import annotations

from functools import lru_cache
from pathlib import Path
from typing import Any, MutableMapping

from .base import Param

try:
    import tomli as toml
except ImportError as e:
    raise ImportError(
        "A TOML package is required for TOML support. "
        "You can install it via `pip install tomli`."
    ) from e

from .base import DispersionMethod

__all__ = ["get_params"]


@lru_cache(maxsize=None)
def _load(method: DispersionMethod) -> dict[str, Any]:
    """Read <method>.toml exactly once."""
    _BASE = Path(__file__).parent
    toml_path = _BASE / f"{method.value}.toml"

    if not toml_path.is_file():
        raise FileNotFoundError(f"TOML file {toml_path} missing.")

    with toml_path.open("rb") as fp:
        return toml.load(fp)


def get_params(
    *,
    method: DispersionMethod | str,
    functional: str | None,
    variant: str | None = None,
    keep_doi: bool = False,
) -> Param:
    """
    Obtain damping parameters for a given functional.

    Parameters
    ----------
    method : DispersionMethod
        Dispersion method (e.g., D3, D4).
    functional : str
        Functional name, case-insensitive.
    variant : str
        Damping variant (e.g., "bj-eeq-atm").
    keep_doi : bool, optional
        If ``True``, keep the DOI field in the output. Defaults to ``False``.

    Raises
    ------
    KeyError
        If functional or dispersion variant is not found in damping
        parameter file.
    """
    method = DispersionMethod(method)
    table = _load(method)

    if functional in (None, "default"):
        default_section = table["default"]

        variant_section = default_section[method.value]
        if variant not in variant_section:
            raise KeyError(
                f"Variant '{variant}' not found in default parameters for "
                f"method={method.value!r}."
            )

        disp_method_section = default_section["parameter"]
    else:
        func_section = table["parameter"]
        if functional not in func_section:
            raise KeyError(
                f"Functional '{functional!r}' not found in damping parameters."
            )

        disp_method_section = func_section[functional.casefold()]
        if method.value not in disp_method_section:
            raise KeyError(
                f"Method '{method}' not found in damping parameters "
                f"for '{functional!r}'."
            )

    # Get default variant if not specified
    if variant is None:
        variant = table["default"][method.value][0]

    variant_section = disp_method_section[method.value]
    if variant not in variant_section:
        raise KeyError(
            f"Variant '{variant}' not found for functional="
            f"{functional!r}, method={method.value!r}."
        )

    block = variant_section[variant]

    out: MutableMapping[str, float | str] = dict(block)
    if not keep_doi and "doi" in out:
        out.pop("doi")

    return Param(**out)  # type: ignore



# This file is part of tad-dftd4.
#
# SPDX-Identifier: Apache-2.0
# Copyright (C) 2024 Grimme Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Damping: Functions
==================

Collections of damping functions including:
- Rational damping (Becke--Johnson)
- Zero damping (Chai--Head-Gordon)
- Modified zero damping
- Optimised power damping
- Becke's Z-damping
"""
from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from functools import wraps

import torch
from tad_mctc import storch
from tad_mctc.typing import Any, Literal, Tensor

from .. import defaults

_MISSING = object()

__all__ = [
    "Damping",
    "RationalDamping",
    "ZeroDamping",
    "MZeroDamping",
    "OptimisedPowerDamping",
]


class Damping(ABC):
    """
    Base interface for damping functions.

    This class defines the interface for damping functions used in DFT-D4
    calculations. It provides a callable interface that takes the order of the
    dispersion interaction, pairwise distances, radii, and parameters as input
    and returns the damping function values.
    """

    radius_type: Literal["r4r2", "rvdw"]
    """Type of radius used in the damping function."""

    doi: str | None = None
    """Digital Object Identifier (DOI) for the damping function."""

    def __init_subclass__(cls, **kw):
        super().__init_subclass__(**kw)

        req_map: dict[int, tuple[str, ...]] = getattr(
            cls, "REQUIRED_BY_ORDER", {}
        )
        if not req_map:  # nothing special declared
            return

        # sanity-check: every declared name must really exist in _f
        sig = inspect.signature(cls._f)
        unknown = {
            name
            for seq in req_map.values()
            for name in seq
            if name not in sig.parameters
        }
        if unknown:
            raise TypeError(
                f"{cls.__name__}.REQUIRED_BY_ORDER refers to "
                f"unknown parameter(s): {', '.join(unknown)}"
            )

        # wrap _f once so every call is validated fast
        original_f = cls._f

        @wraps(original_f)
        def _checked_f(self, *args, **kwargs):
            bound = sig.bind(self, *args, **kwargs)
            bound.apply_defaults()

            order_val = bound.arguments["order"]
            required = req_map.get(order_val, ())

            missing = [
                name
                for name in required
                if bound.arguments.get(name, _MISSING) in (None, _MISSING)
            ]
            if missing:
                raise TypeError(
                    f"{cls.__name__} (order {order_val}) "
                    f"requires keyword(s): {', '.join(missing)}"
                )
            return original_f(self, *args, **kwargs)

        cls._f = _checked_f

    def __call__(
        self,
        distances: Tensor,
        radii: Tensor,
        order: int,
        only_damping: bool = False,
        *,
        a1: Tensor | float | int | None = None,
        a2: Tensor | float | int | None = None,
        s6: Tensor | float | int | None = defaults.S6,
        rs6: Tensor | float | int | None = None,
        s8: Tensor | float | int | None = None,
        rs8: Tensor | float | int | None = None,
        s9: Tensor | float | int | None = defaults.S9,
        rs9: Tensor | float | int | None = defaults.RS9,
        s10: Tensor | float | int | None = None,
        alp: Tensor | float | int | None = defaults.ALP,
        bet: Tensor | float | int | None = None,
        doi: str | None = None,
    ) -> Tensor:
        self.doi = doi

        # Do not skip the cases where the parameter value is 0.0, as this leads
        # to issues for gradient calculations.

        if order == 6 and (s6 is None and rs6 is None):
            return torch.zeros_like(distances)

        if order == 8 and (s8 is None and rs8 is None):
            return torch.zeros_like(distances)

        if order == 9 and (s9 is None and rs9 is None):
            return torch.zeros_like(distances)

        if order == 10 and s10 is None:
            return torch.zeros_like(distances)

        return self._f(
            distances,
            radii,
            order,
            a1=a1,
            a2=a2,
            s6=s6,
            rs6=rs6,
            s8=s8,
            rs8=rs8,
            s9=s9,
            rs9=rs9,
            s10=s10,
            alp=alp,
            bet=bet,
            only_damping=only_damping,
        )

    @abstractmethod
    def _f(
        self,
        distances: Tensor,
        radii: Tensor,
        order: int,
        *,
        a1: Tensor | float | int | None = None,
        a2: Tensor | float | int | None = None,
        s6: Tensor | float | int | None = None,
        rs6: Tensor | float | int | None = None,
        s8: Tensor | float | int | None = None,
        rs8: Tensor | float | int | None = None,
        s9: Tensor | float | int | None = None,
        rs9: Tensor | float | int | None = None,
        s10: Tensor | float | int | None = None,
        alp: Tensor | float | int | None = None,
        bet: Tensor | float | int | None = None,
        only_damping: bool = False,
    ) -> Tensor:
        """
        Calculate the damping function values.

        Parameters
        ----------
        distances : Tensor
            Pairwise distances between atoms in the system.
            (shape: ``(..., nat, nat)``).
        radii : Tensor
            Critical radii for all atom pairs (shape: ``(..., nat, nat)``).
        order : int
            Order of the dispersion interaction, e.g.
            6 for dipole-dipole, 8 for dipole-quadrupole and so on.
        only_damping : bool
            If True, return only the damping function values without division by
            the distance raised to the power of the order.

        Returns
        -------
        Tensor
            Values of the damping function.
        """

    def __eq__(self, other: Any) -> bool:
        """
        Check for equality between two Damping instances.

        Two damping instances are considered equal if they are of the exact
        same type. The specific damping behavior is defined by the class
        itself, and they don't hold any state that would differentiate
        instances of the same class.

        Parameters
        ----------
        other : Any
            The object to compare with.

        Returns
        -------
        bool
            ``True`` if the objects are of the same class, ``False`` otherwise.
        """
        if not isinstance(other, Damping):
            return False
        return self.__class__ is other.__class__


# concrete damping models


class RationalDamping(Damping):
    r"""
    Becke--Johnson rational damping.

    .. math::

        f^n_{\text{damp}}\left(R_0^{\text{AB}}\right) =
        \dfrac{R^n_{\text{AB}}}{R^n_{\text{AB}} +
        \left( a_1 R_0^{\text{AB}} + a_2 \right)^n}
    """

    radius_type = "r4r2"
    """Expectation value of r⁴/r² for the atoms in the system."""

    REQUIRED_BY_ORDER = {
        6: ("a1", "a2"),
        8: ("a1", "a2"),
        10: ("a1", "a2"),
    }
    """Required parameters for the damping function."""

    def _f(
        self,
        distances: Tensor,
        radii: Tensor,
        order: int,
        *,
        a1: Tensor | float | int | None = defaults.A1,
        a2: Tensor | float | int | None = defaults.A2,
        s6: Tensor | float | int | None = None,
        rs6: Tensor | float | int | None = None,
        s8: Tensor | float | int | None = None,
        rs8: Tensor | float | int | None = None,
        s9: Tensor | float | int | None = None,
        rs9: Tensor | float | int | None = None,
        s10: Tensor | float | int | None = None,
        alp: Tensor | float | int | None = None,
        bet: Tensor | float | int | None = None,
        only_damping: bool = False,
    ) -> Tensor:
        """
        Rational damped dispersion interaction between pairs.

        Parameters
        ----------
        distances : Tensor
            Pairwise distances between atoms in the system.
            (shape: ``(..., nat, nat)``).
        radii : Tensor
            Critical radii for all atom pairs (shape: ``(..., nat, nat)``).
        param : Param
            DFT-D4 damping parameters.
        order : int
            Order of the dispersion interaction, e.g.
            6 for dipole-dipole, 8 for dipole-quadrupole and so on.

        Returns
        -------
        Tensor
            Values of the damping function.
        """
        assert a1 is not None and a2 is not None

        radius = a1 * torch.sqrt(radii) + a2
        return 1.0 / (distances.pow(order) + radius.pow(order))


class ZeroDamping(Damping):
    r"""
    Zero damping (also known as Chai--Head-Gordon damping).

    .. math::

        f^n_{\text{damp}}\left(R^{\text{AB}}\right) =
        \dfrac{1}{1 + 6 \left( \dfrac{ R^{\text{AB}} }{ R_0^{\text{AB}} } \right)^{-a}} =
        \dfrac{1}{1 + 6 \left( \dfrac{ R_0^{\text{AB}} }{ R^{\text{AB}} } \right)^{a}}
    """

    radius_type = "rvdw"
    """Pair-wise van-der-Waals radii."""

    REQUIRED_BY_ORDER = {
        6: ("rs6", "alp"),
        8: ("rs8", "alp"),
        9: ("rs9", "alp"),
    }
    """Required parameters for the damping function."""

    def _f(
        self,
        distances: Tensor,
        radii: Tensor,
        order: int,
        *,
        a1: Tensor | float | int | None = None,
        a2: Tensor | float | int | None = None,
        s6: Tensor | float | int | None = None,
        rs6: Tensor | float | int | None = defaults.RS6,
        s8: Tensor | float | int | None = None,
        rs8: Tensor | float | int | None = defaults.RS8,
        s9: Tensor | float | int | None = None,
        rs9: Tensor | float | int | None = defaults.RS9,
        s10: Tensor | float | int | None = None,
        alp: Tensor | float | int | None = defaults.ALP,
        bet: Tensor | float | int | None = None,
        only_damping: bool = False,
    ) -> Tensor:
        assert alp is not None

        if order not in (6, 8, 9):
            raise ValueError(
                "Zero-damping is only implemented for order 6, 8, and 9. "
                f"Got {order} instead."
            )

        if order == 6:
            assert rs6 is not None
            alp_n = alp
            rs = rs6
        elif order == 8:
            assert rs8 is not None
            alp_n = alp + 2.0
            rs = rs8
        elif order == 9:
            assert rs9 is not None
            alp_n = alp
            rs = rs9

        # rs6 * r0ij / r1 = rs6 * rvdw(i,j) / sqrt(r2)
        t_n = rs * storch.divide(distances, radii)

        f_n = 1.0 / (1.0 + 6.0 * t_n**alp_n)

        if only_damping is True:
            return f_n

        # f6 / r6
        return storch.divide(f_n, distances**order)


class MZeroDamping(Damping):
    """Modified zero damping."""

    radius_type = "rvdw"
    """Pair-wise van-der-Waals radii."""

    REQUIRED_BY_ORDER = {
        6: ("rs6", "alp", "bet"),
        8: ("rs8", "alp", "bet"),
    }
    """Required parameters for the damping function."""

    def _f(
        self,
        distances: Tensor,
        radii: Tensor,
        order: int,
        *,
        a1: Tensor | float | int | None = None,
        a2: Tensor | float | int | None = None,
        s6: Tensor | float | int | None = None,
        rs6: Tensor | float | int | None = defaults.RS6,
        s8: Tensor | float | int | None = None,
        rs8: Tensor | float | int | None = defaults.RS8,
        s9: Tensor | float | int | None = None,
        rs9: Tensor | float | int | None = None,
        s10: Tensor | float | int | None = None,
        alp: Tensor | float | int | None = defaults.ALP,
        bet: Tensor | float | int | None = defaults.BET,
        only_damping: bool = False,
    ) -> Tensor:
        assert alp is not None
        assert bet is not None

        if order not in (6, 8):
            raise ValueError(
                "Zero-damping is only implemented for order 6 and 8."
            )

        if order == 6:
            assert rs6 is not None
            alp_n = alp
            rs = rs6
        elif order == 8:
            assert rs8 is not None
            alp_n = alp + 2.0
            rs = rs8

        # t6 = (r1 / (rs6*r0ij) + bet*r0ij)**(-alp6)
        t_n = distances / (rs * radii) + bet * radii

        # f6 / r6
        fn = 1.0 / (1.0 + 6.0 * t_n ** (-alp_n))
        return storch.divide(fn, distances**order)


class OptimisedPowerDamping(Damping):
    """Optimised-power damping."""

    radius_type = "r4r2"
    """Expectation value of r⁴/r² for the atoms in the system."""

    REQUIRED_BY_ORDER = {
        6: ("a1", "a2", "bet"),
        8: ("a1", "a2", "bet"),
    }
    """Required parameters for the damping function."""

    def _f(
        self,
        distances: Tensor,
        radii: Tensor,
        order: int,
        *,
        a1: Tensor | float | int | None = defaults.A1,
        a2: Tensor | float | int | None = defaults.A2,
        s6: Tensor | float | int | None = None,
        rs6: Tensor | float | int | None = None,
        s8: Tensor | float | int | None = None,
        rs8: Tensor | float | int | None = None,
        s9: Tensor | float | int | None = None,
        rs9: Tensor | float | int | None = None,
        s10: Tensor | float | int | None = None,
        alp: Tensor | float | int | None = None,
        bet: Tensor | float | int | None = defaults.BET,
        only_damping: bool = False,
    ) -> Tensor:
        assert a1 is not None
        assert a2 is not None
        assert bet is not None

        if order not in (6, 8):
            raise ValueError(
                "OP-damping is only implemented for order 6 and 8."
            )

        radius = a1 * torch.sqrt(radii) + a2
        ab = radius**bet

        # r2**(bet * 0.5)
        rb = distances**bet

        # t6 = rb / (rb * r2**3 + ab * r0ij**6)
        return rb / (rb * distances**order + ab * distances**order)


class ZDamping(Damping):
    """Becke's Z-damping."""
